
import java.io.File
import java.time.Month
import java.time.MonthDay
import java.time.Year
import java.util.*
import kotlin.collections.ArrayList
import kotlin.collections.HashMap
import kotlin.reflect.KClass

/*
generator
 */

//
// constants
const val LIFE_MIN_AGE = 40
const val LIFE_MAX_AGE = 80

const val DYNASTY_ID_OFFSET = 10000
const val CHARACTER_ID_OFFSET = 600000

//
// the code
open class Counter(initialValue: Int) {
    private var value: Int = initialValue

    fun next(): Int {
        val result = value
        value++
        return result
    }
}

object CharacterIdCounter : Counter(CHARACTER_ID_OFFSET)
object DynastyIdCounter : Counter(DYNASTY_ID_OFFSET)

interface HasPdxString {
    fun toPdxString(indent: Int): String
}

abstract class HasId(val id: Int) : HasPdxString {
}

class Dynasty(id: Int, val name: String, val culture: Culture) : HasId(id) {

    var useExtraCoa = false

    private val coaTemplate =
            """    coat_of_arms = {
        template = 0
        layer = {
            texture = 0
            texture_internal = 1
            emblem = 1
            color = 3
            color = 3
            color = 3
        }
    }
"""

    override fun toPdxString(indent: Int): String {
        val sb = StringBuilder()

        sb.append("$id=\n")
        sb.append("{\n")

        indent(sb, 1)
        sb.append("name = \"$name\"\n")

        indent(sb, 1)
        sb.append("culture = ${culture.pdxName}\n")

        if (useExtraCoa) {
            sb.append(coaTemplate)
        }

        sb.append("}\n")

        return sb.toString()
    }
}

abstract class AbstractStore<T : HasId> : HasPdxString {
    protected val elements: MutableList<T> = ArrayList()


    fun getElementsCopy(): List<T> {
        return ArrayList(elements)
    }

    protected open fun addElement(elem: T) {
        elements.add(elem)
    }

    override fun toString(): String {
        return "Store with ${elements.size} elements"
    }

    override fun toPdxString(indent: Int): String {
        val minId = getFirstId()
        val maxId = getLastId()
        val sb = StringBuilder()
        sb.append("#\n")
        sb.append("# Generated file\n")
        sb.append("#\n")
        sb.append("# IDs used:\n")
        sb.append("# $minId - $maxId\n")
        sb.append("#\n")

        for (c in elements) {
            sb.append(c.toPdxString(indent))
        }
        return sb.toString()
    }

    fun getFirstId(): Int {
        var result: Int = Integer.MAX_VALUE
        for (e in elements) {
            result = Math.min(e.id, result)
        }
        return result
    }

    fun getLastId(): Int {
        var result = 0
        for (e in elements) {
            result = Math.max(e.id, result)
        }
        return result
    }

}

class DynastyStore(val filename: String) : AbstractStore<Dynasty>() {

    fun addDynasty(name: String, culture: Culture): Dynasty {
        val result = Dynasty(DynastyIdCounter.next(), name, culture)

        addElement(result)
        return result
    }

}

class ParadoxDate : Comparable<ParadoxDate> {
    val year: Year
    val month: Month
    val day: MonthDay

    constructor(year: Year, month: Month, day: MonthDay) {
        this.year = year
        this.month = month
        this.day = day
    }

    constructor(year: Int, month: Int, day: Int) : this(
            Year.of(year),
            Month.of(month),
            MonthDay.of(Month.of(month), day)
    )

    constructor(year: Int, month: Int) : this(year, month, 1)

    constructor(year: Int) : this(year, 1)

    override fun toString(): String {
        return "${year.toString()}.${month.value}.${day.dayOfMonth}"
    }

    fun addYears(years: Int): ParadoxDate {
        val newYear = year.plusYears(years.toLong())
        return ParadoxDate(newYear, month, day)
    }

    override operator fun compareTo(other: ParadoxDate): Int {
        return when {
            year != other.year -> year.compareTo(other.year)
            month != other.month -> month.compareTo(other.month)
            day != other.day -> day.compareTo(other.day)
            else -> 0
        }
    }

    override fun equals(other: Any?): Boolean {
        if (other !is ParadoxDate) {
            return false
        }
        return compareTo(other) == 0
    }

    override fun hashCode(): Int {
        var result = year.hashCode()
        result = 31 * result + month.hashCode()
        result = 31 * result + day.hashCode()
        return result
    }
}

interface Event : HasPdxString {

}

class BirthEvent : Event {
    override fun toPdxString(indent: Int): String {
        val sb = StringBuilder(indent + 10)
        indent(sb, indent)
        sb.append("birth=yes")
        return sb.toString()
    }
}

class DeathEvent : Event {
    override fun toPdxString(indent: Int): String {
        val sb = StringBuilder(indent + 10)
        indent(sb, indent)
        sb.append("death=yes")
        return sb.toString()
    }
}

class AddSpouseEvent(val spouse: Character) : Event {

    override fun toPdxString(indent: Int): String {
        val sb = StringBuilder(indent + 10)
        indent(sb, indent)
        sb.append("add_spouse=${spouse.id}")
        return sb.toString()
    }
}

class HolderEvent(val character: Character) : Event {
    override fun toPdxString(indent: Int): String {
        val sb = StringBuilder(indent + 10)
        indent(sb, indent)
        sb.append("holder=${character.id}")
        return sb.toString()
    }
}

class LiegeEvent(val liege: TitleHistory) : Event {
    override fun toPdxString(indent: Int): String {
        val sb = StringBuilder(indent + 10)
        indent(sb, indent)
        sb.append("liege=${liege.name}")
        return sb.toString()
    }
}

class DeJureLiegeEvent(val liege: TitleHistory) : Event {
    override fun toPdxString(indent: Int): String {
        val sb = StringBuilder(indent + 10)
        indent(sb, indent)
        sb.append("de_jure_liege=${liege.name}")
        return sb.toString()
    }
}

class RawDeJureLiegeEvent(val liege: String) : Event {
    override fun toPdxString(indent: Int): String {
        val sb = StringBuilder(indent + 10)
        indent(sb, indent)
        sb.append("de_jure_liege=$liege")
        return sb.toString()
    }
}

class TitleNameEvent(val nameLiteral: String) : Event {
    override fun toPdxString(indent: Int): String {
        val sb = StringBuilder()
        indent(sb, indent)
        sb.append("name = $nameLiteral")
        return sb.toString()
    }
}

class TitleAdjectiveEvent(val adjectiveLiteral: String) : Event {
    override fun toPdxString(indent: Int): String {
        val sb = StringBuilder()
        indent(sb, indent)
        sb.append("adjective = $adjectiveLiteral")
        return sb.toString()
    }
}

class CombinedEvent(val events: List<Event>) : Event {
    override fun toPdxString(indent: Int): String {
        val sb = StringBuilder(indent + 10)
        for (e in events) {
            sb.append(e.toPdxString(indent))
            if (events.indexOf(e) != events.size - 1) {
                sb.append("\n")
            }
        }
        return sb.toString()
    }
}

open class EventStore : HasPdxString {
    private val events: MutableMap<ParadoxDate, Event> = hashMapOf()

    fun addEvent(date: ParadoxDate, eventToAdd: Event) {
        if (!events.containsKey(date)) {
            events[date] = eventToAdd
            return
        }

        val oldEvent = events[date]!!

        val newEvent: CombinedEvent = if (oldEvent is CombinedEvent) {
            val newEventList = ArrayList(oldEvent.events)
            newEventList.add(eventToAdd)
            CombinedEvent(newEventList)
        } else {
            CombinedEvent(listOf(oldEvent, eventToAdd))
        }

        events[date] = newEvent
    }

    fun getSortedEvents(): SortedMap<ParadoxDate, Event> {
        return events.toSortedMap()
    }

    fun getFirstDate(): ParadoxDate {
        return events.toSortedMap().firstKey()
    }

    fun getEventDate(eventType: KClass<out Event>): ParadoxDate? {

        for ((k, v) in events) {
            if (v::class == eventType) {
                return k
            }
        }

        return null
    }

    override fun toString(): String {
        return "EventStore with ${events.size} events"
    }

    override fun toPdxString(indent: Int): String {
        val sb = StringBuilder()

        val sortedEvents = getSortedEvents()


        for ((k, v) in sortedEvents) {

            indent(sb, indent)
            sb.append("$k={\n")

            indent(sb, indent)
            // TODO properly indent combined events
            sb.append(v.toPdxString(indent + 1)).append("\n")

            indent(sb, indent)
            sb.append("}\n")
        }

        return sb.toString()
    }
}

class CharacterEventStore() : EventStore() {

    fun getBirth(): ParadoxDate {
        val result = getEventDate(BirthEvent::class) ?: throw Exception("character doesnt have a birth")
        return result
    }

    fun getDeath(): ParadoxDate {
        val result = getEventDate(DeathEvent::class) ?: throw Exception("character doesnt have a death")
        return result
    }

}

class Character(id: Int, val name: String, val culture: Culture, val religion: Religion, birth: ParadoxDate) : HasId(id) {

    val eventStore: CharacterEventStore = CharacterEventStore()

    /**
     * can be null to allow wives without dynasty
     */
    var dynasty: Dynasty? = null

    var father: Character? = null
    var mother: Character? = null
    var heir: Character? = null
    var firstKidYearBorn: Int? = null
    var ruleStart: ParadoxDate? = null

    var female: Boolean = false

    val children: MutableList<Character> = ArrayList()

    init {
        eventStore.addEvent(birth, BirthEvent())
    }

    companion object {

        fun marry(husband: Character, wife: Character) {
            val husbandMin = husband.getBirth().addYears(16)
            val wifeMin = wife.getBirth().addYears(16)
            val marriage = if (husbandMin < wifeMin) wifeMin else husbandMin

            val husbandSpouseEvent = AddSpouseEvent(wife)
            husband.eventStore.addEvent(marriage, husbandSpouseEvent)

            val wifeSpouseEvent = AddSpouseEvent(husband)
            wife.eventStore.addEvent(marriage, wifeSpouseEvent)

        }
    }

    fun getBirth(): ParadoxDate {
        return eventStore.getBirth()
    }

    fun getDeath(): ParadoxDate {
        return eventStore.getDeath()
    }

    override fun toString(): String {
        return "Character $id $name dynasty ${dynasty?.name}"
    }

    override fun toPdxString(indent: Int): String {
        val sb = StringBuilder()
        val localIndent = 1

        sb.append("$id={\n")

        indent(sb, indent + localIndent)
        sb.append("name=\"$name\"\n")

        indent(sb, indent + localIndent)
        sb.append("religion=${religion.pdxName}\n")

        indent(sb, indent + localIndent)
        sb.append("culture=${culture.pdxName}\n")

        if (dynasty != null) {
            indent(sb, indent + localIndent)
            sb.append("dynasty=${dynasty!!.id}\n")
        }

        if (father != null) {
            indent(sb, indent + localIndent)
            sb.append("father=${father!!.id}\n")
        }

        if (mother != null) {
            indent(sb, indent + localIndent)
            sb.append("mother=${mother!!.id}\n")
        }

        if (female) {
            indent(sb, indent + localIndent)
            sb.append("female=yes\n")
        }

        if (ruleStart != null) {
            indent(sb, indent + localIndent)
            sb.append("# ruleStart = $ruleStart\n")
        }

        if (heir != null) {
            indent(sb, indent + localIndent)
            sb.append("# heir = ${heir!!.id}\n")
        }

        sb.append(eventStore.toPdxString(indent + localIndent))

        sb.append("}\n")

        return sb.toString()
    }
}

class CharacterFactory {
    companion object {
        fun buildCharacter(name: String, culture: Culture, religion: Religion, birth: ParadoxDate): Character {
            val result = Character(CharacterIdCounter.next(), name, culture, religion, birth)
            return result
        }
    }
}

class CharacterStore(
        val filename: String,
        dynasty: Dynasty,
        startYear: Int,
        endYear: Int,
        religion: Religion
) : AbstractStore<Character>() {

    init {
        createCharacters(dynasty, startYear, endYear, religion)
    }

    private fun addCharacter(name: String, culture: Culture, religion: Religion, birth: ParadoxDate): Character {
        val result = CharacterFactory.buildCharacter(name, culture, religion, birth)
        addElement(result)
        return result
    }

    fun getRootCharacter(): Character? {
        var root: Character? = null
        for (c in elements) {
            if (root == null) {
                root = c
                continue
            }

            if (c.female) continue
            if (c.father != null) continue

            // males without a father have to be root
            root = c
        }
        return root
    }

    fun createCharacters(dynasty: Dynasty, startYear: Int, endYear: Int, religion: Religion) {

        val rnd = Random()

        fun deathYear(birthYear: Int): Int {
            val liveMin = LIFE_MIN_AGE
            val liveMax = LIFE_MAX_AGE
            val lifeVariance = liveMax - liveMin
            val liveDelta = rnd.nextInt(lifeVariance)
            return birthYear + liveMin + liveDelta
        }

        fun createRoot(): Character {
            val name = randomName(dynasty.culture.maleNames)
            val culture = dynasty.culture

            var yearNow = startYear

            // now 16-36
            val ageNow = 10 + rnd.nextInt(20)
            val birthYear = yearNow - ageNow
            val birthDate = ParadoxDate(birthYear)

            // kids
            // for now, 1
            val kidsAgeDelta = Math.max(20 - ageNow, 0)
            yearNow += kidsAgeDelta

            // 1st kid
            val firstKidYearBorn = yearNow

            val deathDate = ParadoxDate(deathYear(birthYear))

            val result = addCharacter(name, culture, religion, birthDate)
            result.dynasty = dynasty

            result.eventStore.addEvent(deathDate, DeathEvent())

            result.firstKidYearBorn = firstKidYearBorn

            result.ruleStart = birthDate.addYears(16)

            return result
        }

        fun createWife(husband: Character): Character {
            val name = randomName(dynasty.culture.femaleNames)
            val culture = husband.culture

            val birthYear: Int
            if (husband.firstKidYearBorn == null) {
                birthYear = husband.getBirth().year.value + 3
            } else {

                // born 1st kid with 17-20
                val bornFirstKidWith = 17 + rnd.nextInt(4)

                birthYear = husband.firstKidYearBorn!! - bornFirstKidWith
            }
            val birthDate = ParadoxDate(birthYear)

            val deathDate = ParadoxDate(deathYear(birthYear))

            val result = addCharacter(name, culture, religion, birthDate)
            result.eventStore.addEvent(deathDate, DeathEvent())

            result.female = true

            Character.marry(husband, result)

            return result
        }

        fun createRuler(father: Character, mother: Character): Character {
            assert(father.firstKidYearBorn != null)
            assert(mother.firstKidYearBorn != null)

            val name = randomName(dynasty.culture.maleNames)
            val culture = father.culture

            val birthYear: Int = father.firstKidYearBorn!!
            val birthDate = ParadoxDate(birthYear)

            // get 1st kid with 17-22
            val getFirstKidWith = 17 + rnd.nextInt(6)
            val firstKidYearBorn = birthYear + getFirstKidWith

            val deathDate = ParadoxDate(deathYear(birthYear))

            val result = addCharacter(name, culture, religion, birthDate)
            result.eventStore.addEvent(deathDate, DeathEvent())
            result.firstKidYearBorn = firstKidYearBorn
            result.dynasty = dynasty

            result.father = father
            result.mother = mother
            father.children.add(result)
            mother.children.add(result)

            father.heir = result
            mother.heir = result

            return result
        }

        val dynastyFounder = createRoot()

        var dynastyHead: Character = dynastyFounder
        while (dynastyHead.getDeath().year.value <= endYear) {
            val wife: Character = createWife(dynastyHead)
            val son: Character = createRuler(dynastyHead, wife)
            dynastyHead = son
        }

        // 2nd pass - set ruleStart
        var currentRuler = dynastyFounder
        while (currentRuler.heir != null) {
            // find heir
            var heir = currentRuler.heir!!
            while (heir.getDeath() <= currentRuler.getDeath() && heir.heir != null) {
                heir = heir.heir!!
            }
            heir.ruleStart = currentRuler.getDeath()
            currentRuler = heir
        }
    }

}

class TitleHistory : HasPdxString {
    val events: EventStore = EventStore()
    val name: String

    constructor(characters: List<Character>, name: String) {
        this.name = name
        for (c in characters) {
            if (c.ruleStart == null) {
                continue
            }

            events.addEvent(c.ruleStart!!, HolderEvent(c))
        }
    }

    fun setLiege(liege: TitleHistory) {
        events.addEvent(events.getFirstDate(), LiegeEvent(liege))
    }

    fun setDeJureLiege(liege: TitleHistory) {
        events.addEvent(events.getFirstDate(), DeJureLiegeEvent(liege))
    }

    fun setDeJureLiegeRaw(liege: String) {
        events.addEvent(events.getFirstDate(), RawDeJureLiegeEvent(liege))
    }

    override fun toString(): String {
        return "TitleHistory with ${events.getSortedEvents().size} events"
    }

    override fun toPdxString(indent: Int): String {
        return events.toPdxString(indent)
    }
}

fun writeToFile(string: String, filename: String) {
    File(filename).bufferedWriter().use { out -> out.write(string) }
}

fun randomName(names: List<String>): String {
    val size = names.size
    val r = Random().nextInt(size)
    return names.get(r)
}


fun indent(sb: StringBuilder, indent: Int) {
    sb.append(getIndentString(indent))
}

fun getIndentString(indent: Int): String {
    val whitePerStep = 4
    fun indentString(): String {
        val sb = StringBuilder(indent * whitePerStep)
        (1..indent * whitePerStep).forEach { sb.append(' ') }
        return sb.toString()
    }

    return when (indent) {
        0 -> ""
        1 -> "    "
        2 -> "        "
        3 -> "            "
        else -> indentString()
    }
}

fun addDuchy(
        dStore: DynastyStore,
        cStores: MutableList<CharacterStore>,
        titles: MutableList<TitleHistory>,
        newDynastyName: String,
        culture: Culture,
        religion: Religion,
        duchyName: String,
        mainCounty: String,
        otherCounties: List<String>,
        startYear: Int,
        endYear: Int,
        liege: TitleHistory?,
        deJureLiege: TitleHistory?,
        strongDuke: Boolean): TitleHistory {

    val dynasty = dStore.addDynasty(newDynastyName, culture)

    val cStoreDuchy = CharacterStore("characters_$duchyName.txt", dynasty, startYear, endYear, religion)
    cStores.add(cStoreDuchy)

    val duchy = TitleHistory(cStoreDuchy.getElementsCopy(), duchyName)
    titles.add(duchy)

    if (liege != null) duchy.setLiege(liege)
    if (deJureLiege != null) duchy.setDeJureLiege(deJureLiege)

    val county = TitleHistory(cStoreDuchy.getElementsCopy(), mainCounty)
    county.setLiege(duchy)
    titles.add(county)

    for (c in otherCounties) {
        val cStoreTmp =
                if (strongDuke) {
                    cStoreDuchy
                } else {
                    val cStore = CharacterStore("characters_$c.txt", dynasty, startYear + 30, endYear, religion)
                    cStores.add(cStore)
                    setParent(cStoreDuchy, cStore)
                    cStore
                }

        val countyTmp = TitleHistory(cStoreTmp.getElementsCopy(), c)
        countyTmp.setLiege(duchy)
        titles.add(countyTmp)
    }

    return duchy
}

enum class Religion(val pdxName: String) {
    JEWISH("jewish"),
    KARAITE("karaite"),
    NESTORIAN("nestorian"),
    ORTHODOX("orthodox"),
    MIAPHYSITE("miaphysite"),
    ZOROASTRIAN("zoroastrian"),
    MAZDAKI("mazdaki"),
    KHURMAZTA("khurmazta"),
    CATHOLIC("catholic"),
    SAMARITAN("samaritan")
}

enum class Culture(val pdxName: String, val maleNames: List<String>, val femaleNames: List<String>) {
    HEBREW("hebrew", NAMES_HEBREW_MALE.split(" "), NAMES_HEBREW_FEMALE.split(" ")),
    MIZRACHI("mizrachi", NAMES_MIZRACHI_MALE.split(" "), NAMES_MIZRACHI_FEMALE.split(" ")),
    ARAMAIC("aramaic", NAMES_ARAMAIC_MALE.split(" "), NAMES_ARAMAIC_FEMALE.split(" ")),
    PHOENICIAN("phoenician", NAMES_PHOENICIAN_MALE.split(" "), NAMES_PHOENICIAN_FEMALE.split(" ")),
    ASSYRIAN("assyrian", NAMES_ASSYRIAN_MALE.split(" "), NAMES_ASSYRIAN_FEMALE.split(" ")),
    GREEK("greek", NAMES_GREEK_MALE.split(" "), NAMES_GREEK_FEMALE.split(" ")),
    ARMENIAN("armenian", NAMES_ARMENIAN_MALE.split(" "), NAMES_ARMENIAN_FEMALE.split(" ")),
    KURDISH("kurdish", NAMES_KURDISH_MALE.split(" "), NAMES_KURDISH_FEMALE.split(" ")),
    PERSIAN("persian", NAMES_PERSIAN_MALE.split(" "), NAMES_PERSIAN_FEMALE.split(" ")),
    BALOCH("baloch", NAMES_BALOCH_MALE.split(" "), NAMES_BALOCH_FEMALE.split(" ")),
    SOGDIAN("sogdian", NAMES_SOGDIAN_MALE.split(" "), NAMES_SOGDIAN_FEMALE.split(" ")),
    EGYPTIAN("coptic", NAMES_EGYPTIAN_MALE.split(" "), NAMES_EGYPTIAN_FEMALE.split(" ")),
    SUEBI("suebi", NAMES_SUEBI_MALE.split(" "), NAMES_SUEBI_FEMALE.split(" ")),
    VISIGOTHIC("visigothic", NAMES_VISIGOTHIC_MALE.split(" "), NAMES_VISIGOTHIC_FEMALE.split(" ")),
    BERBER("maghreb_arabic", NAMES_BERBER_MALE.split(" "), NAMES_BERBER_FEMALE.split(" "))
}

fun writeFiles(dStore: DynastyStore, cStores: List<CharacterStore>, titles: List<TitleHistory>) {
    writeToFile(dStore.toPdxString(0), "gen/dynasties/" + dStore.filename)
    println("wrote dynasty file to ${dStore.filename}")

    for (cStore in cStores) {
        writeToFile(cStore.toPdxString(0), "gen/characters/" + cStore.filename)
        println("wrote characters file to ${cStore.filename}")
    }

    for (title in titles) {
        writeToFile(title.toPdxString(0), "gen/titlehistory/${title.name}.txt")
        println("wrote titleholders file to ${title.name}")
    }

    writeBookmarks(titles)
}

fun writeBookmarks(titles: List<TitleHistory>) {
    val bookmarkTitles = listOf("k_israel", "d_damascus", "d_jibal")
    val bookmarkCharactersFilename = "gen/01_bookmarks.txt"
    val bookmarkLocalizationFilename = "gen/BarKokhba_bookmarks.csv"

    fun getHolderEvent(event: Event): HolderEvent? {
        if (event is CombinedEvent) {
            val combinedEvent = event
            for (e in combinedEvent.events) {
                if (e is HolderEvent) {
                    return e
                }
            }
        }

        if (event is HolderEvent) return event

        return null
    }

    /**
     * possibly adds found ruler to rulers list
     */
    fun findRulerForDate(title: TitleHistory, date: ParadoxDate, rulers: MutableMap<String, Character>) {
        val events = title.events.getSortedEvents()

        var currentHolder: Character? = null

        for ((eventDate, event) in events) {
            val holderEvent: HolderEvent = getHolderEvent(event) ?: continue

            if (eventDate > date) break

            currentHolder = holderEvent.character
        }

        if (currentHolder != null) rulers.put(title.name, currentHolder)
    }

    fun buildBookmarkCharacterString(character: Character, titleName: String, date: ParadoxDate): String {
        val age: Int = date.year.value - character.getBirth().year.value

        val prettyName: String = if (titleName == "k_israel") {
            "JUDEA"
        } else {
            titleName
        }
        val sb = StringBuilder()
        sb.append("        #\n")
        sb.append("        # ${character.name} ${character.dynasty!!.name}\n")
        sb.append("        # Ruler of $titleName\n")
        sb.append("        selectable_character = {\n")
        sb.append("            id = ${character.id} \n")
        sb.append("            age = $age\n")
        sb.append("            name = ERA_CHAR_NAME_${character.id}\n")
        sb.append("            title = $titleName\n")
        sb.append("            title_name = $prettyName\n")
        sb.append("            coa_dynasty = yes\n")
        sb.append("            character = {\n")
        sb.append("                dna = TODO\n")
        sb.append("                properties = TODO\n")
        sb.append("                religion=${character.religion.pdxName}\n")
        sb.append("                culture=${character.culture.pdxName}\n")
        sb.append("                government=feudal_government\n")
        sb.append("                dynasty=${character.dynasty!!.id}\n")
        sb.append("            }\n")
        sb.append("        }\n")

        return sb.toString()
    }

    fun buildLocalizationStrings(character: Character, title: String): String {
        val sb = StringBuilder()
        val titleLevel: String = when (title) {
            "k_israel" -> "King"
            "d_damascus" -> "King"
            "d_jibal" -> "Satrab"
            else -> "TODO"
        }

        val printedTitle: String = when (title) {
            "k_israel" -> "Judea"
            "d_damascus" -> "Aram"
            "d_jibal" -> "Jibal"
            else -> title
        }
        sb.append("ERA_CHAR_NAME_${character.id};${character.name};;;;;;;;;;;;;x\n")
        sb.append("ERA_CHAR_INFO_${character.id};$titleLevel ${character.name} of $printedTitle;;;;;;;;;;;;;x\n")
        return sb.toString()
    }

    val bookmarkCharacters: MutableMap<String, Character> = HashMap()
    for (title in titles) {
        if (title.name !in bookmarkTitles) continue

        // find 769 ruler
        val start769 = ParadoxDate(769, 1, 1)
        findRulerForDate(title, start769, bookmarkCharacters)
    }

    val sbBookmarkCharacters = StringBuilder()
    val sbBookmarkLocalization = StringBuilder()

    sbBookmarkLocalization.append("#CODE;ENGLISH;FRENCH;GERMAN;;SPANISH;;;;;;;;;x\n")
    for ((title, character) in bookmarkCharacters) {
        sbBookmarkCharacters.append(buildBookmarkCharacterString(character, title, ParadoxDate(769, 1, 1)))
        sbBookmarkLocalization.append(buildLocalizationStrings(character, title))
    }
    sbBookmarkLocalization.append("#CODE;ENGLISH;FRENCH;GERMAN;;SPANISH;;;;;;;;;x\n")

//    println(sbBookmarkCharacters)
//    println(sbBookmarkLocalization)

    writeToFile(sbBookmarkCharacters.toString(), bookmarkCharactersFilename)
    println("wrote bookmark character templates to $bookmarkCharactersFilename")

    writeToFile(sbBookmarkLocalization.toString(), bookmarkLocalizationFilename)
    println("wrote bookmark localization file to $bookmarkLocalizationFilename")

}

fun setParent(parent: CharacterStore, child: CharacterStore) {
    val rootChild: Character = if (child.getRootCharacter() != null) {
        child.getRootCharacter()!!
    } else {
        return
    }

    var possibleParent: Character = if (parent.getRootCharacter() != null) {
        parent.getRootCharacter()!!
    } else {
        return
    }

    fun canBeParent(parent: Character, child: Character): Boolean {
        return parent.getBirth().addYears(16) < child.getBirth()
                && parent.getDeath().addYears(-1) > child.getBirth()
    }

    while (!canBeParent(possibleParent, rootChild) && possibleParent.heir != null) {
        possibleParent = possibleParent.heir!!
    }

    if (!canBeParent(possibleParent, rootChild)) {
        return
    }

    rootChild.father = possibleParent
    possibleParent.children.add(rootChild)
}

fun doYourWorkOnlyKokhbaIsrael() {
    val START_YEAR = 136
    val END_YEAR = 1337

    val OUTFILE_DYNASTY = "dynasties.txt"

    val dStore = DynastyStore(OUTFILE_DYNASTY)
    val titles = ArrayList<TitleHistory>()
    val cStores = ArrayList<CharacterStore>()

    fun createKingdom(
            cStoreMainDynasty: CharacterStore,
            title: String,
            mainDuchyName: String,
            counties: List<String>
    ): TitleHistory {

        val kingdom = TitleHistory(cStoreMainDynasty.getElementsCopy(), title)
        titles.add(kingdom)

        val mainDuchy = TitleHistory(cStoreMainDynasty.getElementsCopy(), mainDuchyName)
        mainDuchy.setLiege(kingdom)
        mainDuchy.setDeJureLiege(kingdom)
        titles.add(mainDuchy)

        for (c in counties) {
            val newCounty = TitleHistory(cStoreMainDynasty.getElementsCopy(), c)
            newCounty.setLiege(mainDuchy)
            titles.add(newCounty)
        }

        return kingdom
    }
    //
    // Kokhba Dynasty
    val dynastyKokhba = dStore.addDynasty("Kokhba", Culture.HEBREW)
    dynastyKokhba.useExtraCoa = true

    val cStoreKokhba = CharacterStore("characters_kokhba.txt", dynastyKokhba, START_YEAR, END_YEAR, Religion.JEWISH)
    cStores.add(cStoreKokhba)

    val counties2 = listOf("c_hebron", "c_jerusalem", "c_acre")
    val kIsrael = createKingdom(cStoreKokhba, "k_israel", "d_jerusalem", counties2)

    kIsrael.setDeJureLiegeRaw("e_levantia")
    val firstKey = kIsrael.events.getSortedEvents().firstKey()
    kIsrael.events.addEvent(firstKey, TitleNameEvent("JUDEA"))
    kIsrael.events.addEvent(firstKey, TitleAdjectiveEvent("JUDEA_ADJ"))

    val cBeersheb = TitleHistory(cStoreKokhba.getElementsCopy(), "c_beersheb")
    cBeersheb.setLiege(kIsrael)
    cBeersheb.setDeJureLiegeRaw("d_jerusalem")
    titles.add(cBeersheb)

    val cNegev = TitleHistory(cStoreKokhba.getElementsCopy(), "c_negev")
    cNegev.setLiege(kIsrael)
    cNegev.setDeJureLiegeRaw("d_jerusalem")
    titles.add(cNegev)

    writeFiles(dStore, cStores, titles)
}

fun doYourWork3() {
    val START_YEAR = 550
    val END_YEAR = 1350

    val OUTFILE_DYNASTY = "dynasties.txt"

    val dStore = DynastyStore(OUTFILE_DYNASTY)
    val titles = ArrayList<TitleHistory>()
    val cStores = ArrayList<CharacterStore>()

    fun addIndependentDuchy(
            newDynastyName: String,
            culture: Culture,
            religion: Religion,
            duchyName: String,
            mainCounty: String,
            counties: List<String>,
            liege: TitleHistory?,
            startYear: Int,
            strongDuke: Boolean): TitleHistory {
        return addDuchy(
                dStore,
                cStores,
                titles,
                newDynastyName,
                culture,
                religion,
                duchyName,
                mainCounty,
                counties,
                startYear,
                END_YEAR,
                liege,
                null,
                strongDuke
        )
    }

    //
    // cyrenaica - berber/karaite
    val dCyrenaica = addIndependentDuchy(
            "Azimid",
            Culture.BERBER,
            Religion.KARAITE,
            "d_cyrenaica",
            "c_cyrenaica",
            listOf("c_tobruk", "c_senoussi", "c_benghazi"),
            null,
            START_YEAR,
            false)
    // dPalmyra.setLiege(kSyria)

    //
    // tripolitania - berber/karaite
    val dTripolitania = addIndependentDuchy(
            "Malzuzi",
            Culture.BERBER,
            Religion.KARAITE,
            "d_tripolitania",
            "c_tripolitana",
            listOf("c_syrte", "c_leptis_magna", "c_djerba"),
            null,
            START_YEAR,
            false)

    //
    // tunis - berber/karaite
    val dTunis = addIndependentDuchy(
            "Kamalid",
            Culture.BERBER,
            Religion.KARAITE,
            "d_tunis",
            "c_tunis",
            listOf("c_bizerte", "c_medjerda", "c_mahdia", "c_kairwan", "c_gabes"),
            null,
            START_YEAR,
            false)

    //
    // kabylia - berber/karaite
    val dKabylia = addIndependentDuchy(
            "Awzalid",
            Culture.BERBER,
            Religion.KARAITE,
            "d_kabylia",
            "c_beni_yanni",
            listOf("c_tell_atlas", "c_bejaija", "c_biskra", "c_annaba", "c_constantine"),
            null,
            START_YEAR,
            false)

    //
    // alger - berber/karaite
    val dAlger = addIndependentDuchy(
            "Rustamid",
            Culture.BERBER,
            Religion.KARAITE,
            "d_alger",
            "c_al_djazair",
            listOf("c_lemdiyya", "c_atlas_mnt", "c_mzab", "c_ouled_nail", "c_biskra"),
            null,
            START_YEAR,
            false)

    //
    // tlemcen - berber/karaite
    val dTlemcen = addIndependentDuchy(
            "Ifrinid",
            Culture.BERBER,
            Religion.KARAITE,
            "d_tlemcen",
            "c_tlemcen",
            listOf("c_orania", "c_snassen", "c_hanyan"),
            null,
            START_YEAR,
            false)

    //
    // fes - berber/karaite
    val dFes = addIndependentDuchy(
            "Midrarid",
            Culture.BERBER,
            Religion.KARAITE,
            "d_fes",
            "c_fes",
            listOf("c_el_rif", "c_figuig", "c_sijilmasa"),
            null,
            START_YEAR,
            false)

    //
    // tangier - berber/karaite
    val dTangier = addIndependentDuchy(
            "Qutid",
            Culture.BERBER,
            Religion.KARAITE,
            "d_tangiers",
            "c_tangiers",
            listOf("c_cebta", "c_infa"),
            null,
            START_YEAR,
            false)

    //
    // marrakech - berber/karaite
    val dMarrakech = addIndependentDuchy(
            "Masmudid",
            Culture.BERBER,
            Religion.KARAITE,
            "d_marrakech",
            "c_marrakech",
            listOf("c_massat", "c_anti-atlas", "c_ifni", "c_tharasset"),
            null,
            START_YEAR,
            false)

    writeFiles(dStore, cStores, titles)
}

fun doYourWork2() {
    val START_YEAR = 135
    val END_YEAR = 650

    val OUTFILE_DYNASTY = "dynasties.txt"

    val dStore = DynastyStore(OUTFILE_DYNASTY)
    val titles = ArrayList<TitleHistory>()
    val cStores = ArrayList<CharacterStore>()

    //
    // Jerusalem
    val dynastyKokhba = dStore.addDynasty("Kokhba", Culture.HEBREW)

    val cStoreKokhba = CharacterStore("characters_kokhba.txt", dynastyKokhba, START_YEAR, END_YEAR, Religion.JEWISH)
    cStores.add(cStoreKokhba)
    val titleOne = TitleHistory(cStoreKokhba.getElementsCopy(), "title_kokhba")
    titles.add(titleOne)

    //
    // samaritans
    val dynastySamaritan = dStore.addDynasty("Samaritan", Culture.HEBREW)

    val cStoreSamaritan = CharacterStore("characters_samaritan.txt", dynastySamaritan, 486, END_YEAR, Religion.SAMARITAN)
    cStores.add(cStoreSamaritan)
    val titleSamaritan = TitleHistory(cStoreSamaritan.getElementsCopy(), "title_samaritan")
    titles.add(titleSamaritan)

    writeFiles(dStore, cStores, titles)
}

fun doYourWork() {
    val START_YEAR = 140
    val END_YEAR = 1337

    val OUTFILE_DYNASTY = "dynasties.txt"

    val dStore = DynastyStore(OUTFILE_DYNASTY)
    val titles = ArrayList<TitleHistory>()
    val cStores = ArrayList<CharacterStore>()

    fun createKingdom(
            cStoreMainDynasty: CharacterStore,
            title: String,
            mainDuchyName: String,
            counties: List<String>
    ): TitleHistory {

        val kingdom = TitleHistory(cStoreMainDynasty.getElementsCopy(), title)
        titles.add(kingdom)

        val mainDuchy = TitleHistory(cStoreMainDynasty.getElementsCopy(), mainDuchyName)
        mainDuchy.setLiege(kingdom)
        mainDuchy.setDeJureLiege(kingdom)
        titles.add(mainDuchy)

        for (c in counties) {
            val newCounty = TitleHistory(cStoreMainDynasty.getElementsCopy(), c)
            newCounty.setLiege(mainDuchy)
            titles.add(newCounty)
        }

        return kingdom
    }
    //
    // Kokhba Dynasty
    val dynastyKokhba = dStore.addDynasty("Kokhba", Culture.HEBREW)
    dynastyKokhba.useExtraCoa = true

    val cStoreKokhba = CharacterStore("characters_kokhba.txt", dynastyKokhba, START_YEAR, END_YEAR, Religion.JEWISH)
    cStores.add(cStoreKokhba)

    val counties2 = listOf("c_hebron", "c_jerusalem", "c_acre")
    val kIsrael = createKingdom(cStoreKokhba, "k_israel", "d_jerusalem", counties2)

    kIsrael.setDeJureLiegeRaw("e_levantia")
    val firstKey = kIsrael.events.getSortedEvents().firstKey()
    kIsrael.events.addEvent(firstKey, TitleNameEvent("JUDEA"))
    kIsrael.events.addEvent(firstKey, TitleAdjectiveEvent("JUDEA_ADJ"))

    val cBeersheb = TitleHistory(cStoreKokhba.getElementsCopy(), "c_beersheb")
    cBeersheb.setLiege(kIsrael)
    cBeersheb.setDeJureLiegeRaw("d_jerusalem")
    titles.add(cBeersheb)

    val cNegev = TitleHistory(cStoreKokhba.getElementsCopy(), "c_negev")
    cNegev.setLiege(kIsrael)
    cNegev.setDeJureLiegeRaw("d_jerusalem")
    titles.add(cNegev)

    //
    // duchy of ascalon
    val dynastyAscalon = dStore.addDynasty("Ashkeloni", Culture.HEBREW)

    val cStoreAscalon = CharacterStore("characters_d_ascalon.txt", dynastyAscalon, START_YEAR + 200, END_YEAR, Religion.JEWISH)
    cStores.add(cStoreAscalon)

    val dAscalon = TitleHistory(cStoreAscalon.getElementsCopy(), "d_ascalon")
    dAscalon.setLiege(kIsrael)
    dAscalon.setDeJureLiege(kIsrael)
    titles.add(dAscalon)

    val cAscalon = TitleHistory(cStoreAscalon.getElementsCopy(), "c_ascalon")
    cAscalon.setLiege(dAscalon)
    titles.add(cAscalon)

    val cStoreDarum = CharacterStore("characters_c_darum.txt", dynastyAscalon, START_YEAR + 250, END_YEAR, Religion.JEWISH)
    cStores.add(cStoreDarum)
    setParent(cStoreAscalon, cStoreDarum)

    val cDarum = TitleHistory(cStoreDarum.getElementsCopy(), "c_darum")
    cDarum.setLiege(dAscalon)
    titles.add(cDarum)

    val cStoreEilat = CharacterStore("characters_c_eilat.txt", dynastyAscalon, START_YEAR + 300, END_YEAR, Religion.JEWISH)
    cStores.add(cStoreEilat)
    setParent(cStoreAscalon, cStoreEilat)

    val cEilat = TitleHistory(cStoreEilat.getElementsCopy(), "c_eilat")
    cEilat.setLiege(kIsrael)
    cEilat.setDeJureLiegeRaw("d_ascalon")
    titles.add(cEilat)

    //
    // duchy of sinai, kokhba dynasty
    val cStoreSinai = CharacterStore("characters_d_sinai.txt", dynastyKokhba, START_YEAR + 200, END_YEAR, Religion.JEWISH)
    cStores.add(cStoreSinai)
    setParent(cStoreKokhba, cStoreSinai)

    val dSinai = TitleHistory(cStoreSinai.getElementsCopy(), "d_sinai")
    dSinai.setLiege(kIsrael)
    dSinai.setDeJureLiege(kIsrael)
    titles.add(dSinai)

    val cFarama = TitleHistory(cStoreSinai.getElementsCopy(), "c_farama")
    cFarama.setLiege(dSinai)
    titles.add(cFarama)

    val countiesSinai = listOf("c_sinai", "c_el-arish")
    for (c in countiesSinai) {
        val cStoreTmp = CharacterStore("characters_$c.txt", dynastyKokhba, START_YEAR + 200, END_YEAR, Religion.JEWISH)
        cStores.add(cStoreTmp)
        setParent(cStoreKokhba, cStoreTmp)

        val newCounty = TitleHistory(cStoreTmp.getElementsCopy(), c)
        newCounty.setLiege(dSinai)
        titles.add(newCounty)
    }

    fun addIndependentDuchy(
            newDynastyName: String,
            culture: Culture,
            religion: Religion,
            duchyName: String,
            mainCounty: String,
            counties: List<String>,
            liege: TitleHistory?,
            startYear: Int,
            strongDuke: Boolean): TitleHistory {
        return addDuchy(
                dStore,
                cStores,
                titles,
                newDynastyName,
                culture,
                religion,
                duchyName,
                mainCounty,
                counties,
                startYear,
                END_YEAR,
                liege,
                null,
                strongDuke
        )
    }

    //
    // another dynasty, duke of galilee
    addDuchy(
            dStore,
            cStores,
            titles,
            "HaGalil",
            Culture.HEBREW,
            Religion.JEWISH,
            "d_galilee",
            "c_tyrus",
            listOf("c_tiberias", "c_safed", "c_beirut"),
            START_YEAR,
            END_YEAR,
            kIsrael,
            kIsrael,
            false)


    //
    // another dynasty, duke of oulrejourdain
    addDuchy(
            dStore,
            cStores,
            titles,
            "Safra",
            Culture.MIZRACHI,
            Religion.JEWISH,
            "d_oultrejourdain",
            "c_madaba",
            listOf("c_kerak", "c_monreal"),
            START_YEAR + 200,
            END_YEAR,
            kIsrael,
            kIsrael,
            false)

    //
    // kingdom of aram
    val dynastyAdonibaalid = dStore.addDynasty("Adonibaalid", Culture.ARAMAIC)
    val cStoreAdonibaalid = CharacterStore("characters_k_syria.txt", dynastyAdonibaalid, START_YEAR + 100, END_YEAR, Religion.NESTORIAN)
    cStores.add(cStoreAdonibaalid)

    val countiesKSyria = listOf("c_damascus", "c_amman", "c_irbid", "c_az_zarqa", "c_al_mafraq")
    val kSyria = createKingdom(cStoreAdonibaalid, "k_syria", "d_damascus", countiesKSyria)

    kSyria.setDeJureLiegeRaw("e_levantia")
    val firstKeyKSyria = kSyria.events.getSortedEvents().firstKey()
    kSyria.events.addEvent(firstKeyKSyria, TitleNameEvent("ARAM"))
    kSyria.events.addEvent(firstKeyKSyria, TitleAdjectiveEvent("ARAM_ADJ"))

    //
    // another dynasty, duke of palmyra
    val dPalmyra = addIndependentDuchy(
            "Philosirid",
            Culture.ARAMAIC,
            Religion.NESTORIAN,
            "d_palmyra",
            "c_palmyra",
            listOf("c_syria", "c_tadmor", "c_druz"),
            null,
            START_YEAR + 50,
            false)
    dPalmyra.setLiege(kSyria)

    //
    // another dynasty, duke of Tripoli
    addIndependentDuchy(
            "Eshbaalid",
            Culture.PHOENICIAN,
            Religion.NESTORIAN,
            "d_tripoli",
            "c_tripoli",
            listOf("c_tortosa", "c_baalbek"),
            null,
            START_YEAR + 30,
            true)

    //
    // another dynasty, duke of antioch
    addIndependentDuchy(
            "Hiramid",
            Culture.ARAMAIC,
            Religion.NESTORIAN,
            "d_antioch",
            "c_antiocheia",
            listOf("c_archa", "c_alexandretta"),
            null,
            START_YEAR + 50,
            false)


    //
    // another dynasty, duke of aleppo
    addIndependentDuchy(
            "Mattani",
            Culture.ARAMAIC,
            Religion.NESTORIAN,
            "d_aleppo",
            "c_aleppo",
            listOf("c_hama", "c_asas", "c_homs"),
            null,
            START_YEAR + 50,
            false)

    //
    // another dynasty, duke of Mudar, independent
    addIndependentDuchy(
            "Denhaid",
            Culture.ASSYRIAN,
            Religion.NESTORIAN,
            "d_mudar",
            "c_rahbah",
            listOf("c_al_bichri", "c_al_jazira"),
            null,
            START_YEAR + 50,
            false)

    //
    // seleukia
    addIndependentDuchy(
            "Melition",
            Culture.GREEK,
            Religion.ORTHODOX,
            "d_armenia_minor",
            "c_tarsos",
            listOf("c_seleukeia", "c_adana", "c_teluch", "c_lykandos"),
            null,
            START_YEAR + 50,
            false)

    //
    // duchy of cyprus
    addIndependentDuchy(
            "Anatolion",
            Culture.GREEK,
            Religion.ORTHODOX,
            "d_cyprus",
            "c_famagusta",
            listOf("c_limisol"),
            null,
            START_YEAR + 50,
            false)

    //
    // duchy of edessa, independent
    addIndependentDuchy(
            "Sevagid",
            Culture.ARMENIAN,
            Religion.MIAPHYSITE,
            "d_edessa",
            "c_edessa",
            listOf("c_tell_bashir", "c_aintab", "c_melitene"),
            null,
            START_YEAR + 300,
            false)

    //
    // duchy of mosul, independent, assyrian/nestorian
    addIndependentDuchy(
            "Pappaid",
            Culture.ASSYRIAN,
            Religion.NESTORIAN,
            "d_mosul",
            "c_mosul",
            listOf("c_nisibin", "c_sinjar", "c_oromieh"),
            null,
            START_YEAR + 300,
            false)

    //
    // duchy of jazira, independent, kurdish/miaphysite
    addIndependentDuchy(
            "Erdewanid",
            Culture.KURDISH,
            Religion.MIAPHYSITE,
            "d_jazira",
            "c_bira",
            listOf("c_amida"),
            null,
            START_YEAR + 300,
            false)

    //
    // duchy of samarra, independent, assyrian/nestorian
    addIndependentDuchy(
            "Chumaid",
            Culture.ASSYRIAN,
            Religion.NESTORIAN,
            "d_samarra",
            "c_euphrates",
            listOf("c_deir", "c_kirkuk"),
            null,
            START_YEAR + 300,
            false)

    //
    // duchy of kurdistan, independent, kurdish/nestorian
    addIndependentDuchy(
            "Ferazid",
            Culture.KURDISH,
            Religion.NESTORIAN,
            "d_kurdistan",
            "c_kurdistan",
            listOf("c_shahrazur"),
            null,
            START_YEAR + 300,
            false)

    //
    // duchy of tabriz, independent, kurdish/zoroastrian
    addIndependentDuchy(
            "Rezaid",
            Culture.KURDISH,
            Religion.ZOROASTRIAN,
            "d_tabriz",
            "c_tabriz",
            // add shirvan since its duchy is occupied by armenia
            listOf("c_maragha", "c_shirvan"),
            null,
            START_YEAR + 300,
            false)

    //
    // duchy of gilan, independent, persian/zoroastrian

    fun createPersianZoroastrianDuchy(
            newDynastyName: String,
            duchyName: String,
            mainCounty: String,
            counties: List<String>,
            strongDuke: Boolean) {
        addIndependentDuchy(
                newDynastyName,
                Culture.PERSIAN,
                Religion.ZOROASTRIAN,
                duchyName,
                mainCounty,
                counties,
                null,
                START_YEAR + 300,
                strongDuke)
    }

    //
    // kingdom of persia
    val dynastyKPersia = dStore.addDynasty("Sadrid", Culture.PERSIAN)
    val cStoreKPersia = CharacterStore("characters_k_persia.txt", dynastyKPersia, START_YEAR + 300, END_YEAR, Religion.ZOROASTRIAN)
    cStores.add(cStoreKPersia)

    val countiesKPersia = listOf("c_esfahan", "c_luristan", "c_hamadan", "c_qom", "c_zanjan_abhar", "c_rayy")
    val kPersia = createKingdom(cStoreKPersia, "k_persia", "d_jibal", countiesKPersia)

    //
    // duchy of mafaza
    val dMafaza = addIndependentDuchy(
            "Kurushid",
            Culture.PERSIAN,
            Religion.ZOROASTRIAN,
            "d_mafaza",
            "c_lut",
            listOf("c_yazd"),
            null,
            START_YEAR + 50,
            false)
    dMafaza.setLiege(kPersia)

    createPersianZoroastrianDuchy(
            "Fathid",
            "d_gilan",
            "c_gilan",
            listOf("c_daylam", "c_qazwin"),
            false)

    //
    // duchy of tabaristan, independent, persian/zoroastrian
    createPersianZoroastrianDuchy(
            "Darabid",
            "d_mazandaran",
            "c_alamut",
            listOf("c_qwivir", "c_tabaristan", "c_gurgan"),
            false)

    //
    // duchy of khozistan, independent, persian/mazdaki
    addIndependentDuchy(
            "Behradid",
            Culture.PERSIAN,
            Religion.MAZDAKI,
            "d_khozistan",
            "c_khozistan",
            // typo in the game files: game file is avhaz but localized sting is ahvaz
            listOf("c_tigris", "c_avhaz"),
            null,
            START_YEAR + 300,
            false)

    //
    // duchy of fars, independent, persian/zoroastrian
    createPersianZoroastrianDuchy(
            "Zahakid",
            "d_fars",
            "c_fars",
            listOf("c_ladistan", "c_shiraz", "c_istakhr", "c_hendjan"),
            false)

    //
    // duchy of kerman, independent, persian/zoroastrian
    createPersianZoroastrianDuchy(
            "Kambizid",
            "d_kerman",
            "c_kerman",
            // add zahedan which is not de jure
            listOf("c_bam", "c_sirjan", "c_sistan", "c_hormuz", "c_zahedan"),
            false)


    //
    // duchy of khorasan, independent, persian/zoroastrian
    createPersianZoroastrianDuchy(
            "Manushihrid",
            "d_khorasan",
            "c_nishapur",
            listOf("c_tus", "c_qohistan", "c_birjand"),
            false)

    //
    // duchy of herat, independent, persian/zoroastrian
    createPersianZoroastrianDuchy(
            "Aminid",
            "d_herat",
            "c_herat",
            listOf("c_kanj_rustaq", "c_badghis"),
            false)

    //
    // duchy of balkh, independent, persian/zoroastrian
    createPersianZoroastrianDuchy(
            "Afsharid",
            "d_balkh",
            "c_balkh",
            listOf("c_guzgan"),
            false)

    //
    // duchy of merv, independent, persian/zoroastrian
    createPersianZoroastrianDuchy(
            "Jahanshahid",
            "d_merv",
            "c_merv",
            listOf("c_amol", "c_sarakhs", "c_konjikata"),
            false)

    //
    // duchy of dihistan, independent, persian/zoroastrian
    createPersianZoroastrianDuchy(
            "Ghobadid",
            "d_dihistan",
            "c_dihistan",
            listOf("c_kara-kum"),
            false)

    //
    // duchy of khiva, independent, persian/zoroastrian
    createPersianZoroastrianDuchy(
            "Kambizid",
            "d_khiva",
            "c_khiva",
            listOf("c_dashhowuz", "c_urgench"),
            true)

    //
    // duchy of samarkand, independent, sogdian/zoroastrian
    addIndependentDuchy(
            "Shahbazid",
            Culture.SOGDIAN,
            Religion.ZOROASTRIAN,
            "d_samarkand",
            "c_samarkand",
            listOf("c_nakhshab", "c_bukhara", "c_oshrusana"),
            null,
            START_YEAR + 300,
            false)

    //
    // duchy of ferghana, independent, sogdian/khurmazta
    addIndependentDuchy(
            "Farghanad",
            Culture.SOGDIAN,
            Religion.KHURMAZTA,
            "d_ferghana",
            "c_fergana",
            listOf("c_khojand", "c_khaylam"),
            null,
            START_YEAR + 300,
            false)

    //
    // duchy of makran, independent, baloch/zoroastrian
    addIndependentDuchy(
            "Xusrewid",
            Culture.BALOCH,
            Religion.ZOROASTRIAN,
            "d_baluchistan",
            "c_makran",
            listOf("c_tis", "c_saravan", "c_chagai", "c_armail"),
            null,
            START_YEAR + 300,
            false)


    fun createEgyptianDuchy(
            newDynastyName: String,
            duchyName: String,
            mainCounty: String,
            counties: List<String>,
            strongDuke: Boolean): TitleHistory {
        return addIndependentDuchy(
                newDynastyName,
                Culture.EGYPTIAN,
                Religion.MIAPHYSITE,
                duchyName,
                mainCounty,
                counties,
                null,
                START_YEAR + 200,
                strongDuke)
    }

    //
    // kingdom of egypt
    val dynastyKEgypt = dStore.addDynasty("Badrid", Culture.EGYPTIAN)

    val cStoreKEgypt = CharacterStore("characters_k_egypt.txt", dynastyKEgypt, START_YEAR + 200, END_YEAR, Religion.MIAPHYSITE)
    cStores.add(cStoreKEgypt)

    val countiesKEgypt = listOf("c_alexandria", "c_gizeh", "c_quattara", "c_al_alamayn")
    val kEgypt = createKingdom(cStoreKEgypt, "k_egypt", "d_alexandria", countiesKEgypt)

    kEgypt.setDeJureLiegeRaw("e_abyssinia")

    //
    // Duchy of Damietta
    val dDamietta = addIndependentDuchy(
            "Ablarionid",
            Culture.EGYPTIAN,
            Religion.MIAPHYSITE,
            "d_damietta",
            "c_delta",
            listOf("c_pelusia", "c_manupura", "c_gabiyaha"),
            null,
            START_YEAR + 250,
            false)
    dDamietta.setLiege(kEgypt)

    //
    // duchy of cairo, independent, egyptian/miaphysite
    createEgyptianDuchy(
            "Biktorid",
            "d_cairo",
            "c_cairo",
            listOf("c_sarqihya"),
            false)

    //
    // duchy of aswan, independent, egyptian/miaphysite
    createEgyptianDuchy(
            "Mehtarid",
            "d_aswan",
            "c_aswan",
            listOf("c_buhairya", "c_quena", "c_asyut"),
            false)

    //
    // duchy of porto, independent, suebi/catholic
    addIndependentDuchy(
            "Eirigunig",
            Culture.SUEBI,
            Religion.CATHOLIC,
            "d_porto",
            "c_porto",
            listOf("c_braganza", "c_castelo_branco", "c_coimbra"),
            null,
            400,
            false)

    fun createVisigothicDuchy(
            newDynastyName: String,
            duchyName: String,
            mainCounty: String,
            counties: List<String>,
            strongDuke: Boolean) {
        addIndependentDuchy(
                newDynastyName,
                Culture.VISIGOTHIC,
                Religion.CATHOLIC,
                duchyName,
                mainCounty,
                counties,
                null,
                350,
                strongDuke)
    }

    createVisigothicDuchy("Agilaric", "d_beja", "c_lisboa", listOf("c_alcacer_do_sal", "c_evora", "c_mertola"), false)
    createVisigothicDuchy("Mauregatic", "d_algarve", "c_faro", listOf("c_silves"), false)
    createVisigothicDuchy("Fafiling", "d_leon", "c_zamora", listOf("c_salamanca", "c_valladolid", "c_soria"), false)
    createVisigothicDuchy("Ordonning", "d_badajoz", "c_badajoz", listOf("c_caceres", "c_plasencia", "c_alcantara"), false)
    createVisigothicDuchy("Diaconing", "d_sevilla", "c_sevilla", listOf("c_algeciras", "c_cadiz", "c_aracena", "c_niebla"), true)
    createVisigothicDuchy("Luping", "d_granada", "c_granada", listOf("c_malaga", "c_almeria"), false)
    createVisigothicDuchy("Mauregatic", "d_cordoba", "c_cordoba", listOf("c_calatrava", "c_la_mancha", "c_jaen"), true)
    createVisigothicDuchy("Petric", "d_toledo", "c_toledo", listOf("c_cuenca", "c_molina"), false)
    createVisigothicDuchy("Nunnic", "d_murcia", "c_murcia", listOf("c_almansa"), false)
    createVisigothicDuchy("Assuric", "d_valencia", "c_valencia", listOf("c_denia", "c_castellon"), false)
    createVisigothicDuchy("Reccarding", "d_aragon", "c_zaragoza", listOf("c_alto_aragon", "c_albarracin", "c_calatayud", "c_navarra", "c_najera"), false)
    createVisigothicDuchy("Johaning", "d_barcelona", "c_barcelona", listOf("c_tarragona", "c_lleida", "c_empuries", "c_urgell"), false)

    //
    // cyrenaica - coptic/miaphysite
    val dCyrenaica = addIndependentDuchy(
            "Azimid",
            Culture.EGYPTIAN,
            Religion.MIAPHYSITE,
            "d_cyrenaica",
            "c_cyrenaica",
            listOf("c_tobruk", "c_senoussi", "c_benghazi"),
            null,
            START_YEAR + 300,
            false)
    // dPalmyra.setLiege(kSyria)

    //
    // syrte - berber/miaphysite
    val dSyrte = addIndependentDuchy(
            "Usmid",
            Culture.BERBER,
            Religion.MIAPHYSITE,
            "d_syrte",
            "c_syrte",
            listOf("c_ajadabiya", "c_tadjrift"),
            null,
            START_YEAR + 300,
            false)

    //
    // tripolitania - phoenician/jewish (berber/jewish pop)
    val dTripolitania = addIndependentDuchy(
            "Malzuzi",
            Culture.PHOENICIAN,
            Religion.JEWISH,
            "d_tripolitania",
            "c_tripolitana",
            listOf("c_leptis_magna", "c_djerba", "c_nafusa", "c_nalut"),
            null,
            START_YEAR + 300,
            false)

    //
    // jerid - phoenician/orthodox (berber/karaite pop)
    val dJerid = addIndependentDuchy(
            "Malak",
            Culture.PHOENICIAN,
            Religion.ORTHODOX,
            "d_jerid",
            "c_gabes",
            listOf("c_suf", "c_gafsa"),
            null,
            START_YEAR + 300,
            false)

    //
    // tunis - phoenician/orthodox
    val dTunis = addIndependentDuchy(
            "Kamalid",
            Culture.PHOENICIAN,
            Religion.ORTHODOX,
            "d_tunis",
            "c_tunis",
            listOf("c_bizerte", "c_medjerda", "c_mahdia", "c_kairwan"),
            null,
            START_YEAR + 300,
            false)

    //
    // kabylia - berber/karaite
    val dKabylia = addIndependentDuchy(
            "Awzalid",
            Culture.BERBER,
            Religion.KARAITE,
            "d_kabylia",
            "c_bejaija",
            listOf("c_tell_atlas", "c_biskra", "c_annaba", "c_constantine", "c_tebessa"),
            null,
            START_YEAR + 300,
            false)

    //
    // mzab - berber/karaite
    val dMzab = addIndependentDuchy(
            "Rasulid",
            Culture.BERBER,
            Religion.KARAITE,
            "d_mzab",
            "c_mzab",
            listOf("c_arigh", "c_tuggurt", "c_ouled_nail"),
            null,
            START_YEAR + 300,
            false)

    //
    // alger - berber/catholic
    val dAlger = addIndependentDuchy(
            "Rustamid",
            Culture.BERBER,
            Religion.CATHOLIC,
            "d_alger",
            "c_al_djazair",
            listOf("c_orania", "c_beni_yanni"),
            null,
            START_YEAR + 300,
            false)

    //
    // tahert - berber/miaphysite
    val dTahert = addIndependentDuchy(
            "Taharid",
            Culture.BERBER,
            Religion.MIAPHYSITE,
            "d_tahert",
            "c_lemdiyya",
            listOf("c_hanyan", "c_atlas_mnt", "c_ashir"),
            null,
            START_YEAR + 300,
            false)

    //
    // tlemcen - berber/karaite
    val dTlemcen = addIndependentDuchy(
            "Ifrinid",
            Culture.BERBER,
            Religion.KARAITE,
            "d_tlemcen",
            "c_tlemcen",
            listOf("c_snassen", "c_figuig"),
            null,
            START_YEAR + 300,
            false)

    //
    // fes - berber/karaite
    val dFes = addIndependentDuchy(
            "Midrarid",
            Culture.BERBER,
            Religion.KARAITE,
            "d_fes",
            "c_fes",
            listOf("c_taza", "c_muluja", "c_meknes", "c_infa"),
            null,
            START_YEAR + 300,
            false)

    //
    // tangier - berber/catholic
    val dTangier = addIndependentDuchy(
            "Qutid",
            Culture.BERBER,
            Religion.CATHOLIC,
            "d_tangiers",
            "c_tangiers",
            listOf("c_cebta", "c_el_rif"),
            null,
            START_YEAR + 300,
            false)

    //
    // marrakech - berber/karaite
    val dMarrakech = addIndependentDuchy(
            "Masmudid",
            Culture.BERBER,
            Religion.KARAITE,
            "d_marrakech",
            "c_marrakech",
            listOf("c_massat", "c_anti-atlas", "c_tadla", "c_tinmallal"),
            null,
            START_YEAR + 300,
            false)

    writeFiles(dStore, cStores, titles)
}

fun main(args: Array<String>) {
    doYourWork()
}
